/*
 * Copyright 2017 Arne Vajhj.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License. 
 */

package dk.vajhoej.vms.call.lib;

import static dk.vajhoej.vms.call.VMS.*;
import static dk.vajhoej.vms.call.lib.VMSCodes.*;

import java.io.UnsupportedEncodingException;
import java.util.HashMap;
import java.util.Map;

import dk.vajhoej.record.Alignment;
import dk.vajhoej.record.FieldType;
import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.Struct;
import dk.vajhoej.record.StructField;
import dk.vajhoej.vms.call.IOSB;
import dk.vajhoej.vms.call.VMSLibrary;
import dk.vajhoej.vms.call.VMS.CharacterString;
import dk.vajhoej.vms.call.VMS.Item32;
import dk.vajhoej.vms.call.VMS.Item32CharacterStringIn;
import dk.vajhoej.vms.call.VMS.ItemList32;
import dk.vajhoej.vms.call.VMS.LongWord;
import dk.vajhoej.vms.call.VMS.Word;

/**
 * Class SYS is a wrapper library for system services (SYS$ functions).
 * 
 * See VMS documentation for semantics.
 */
public class SYS extends VMSLibrary {
	/**
	 * Class ACMSB represents an ACME status block.
	 */
	@Struct(alignment=Alignment.PACKED)
	public static class ACMESB {
		@StructField(n=0,type=FieldType.INT4)
		private int status;
		@StructField(n=1,type=FieldType.INT4)
		private int status2;
		@StructField(n=2,type=FieldType.INT4)
		private int acmeId;
		@StructField(n=3,type=FieldType.INT4)
		private int acmeStatus;
		/**
		 * Construct new instance.
		 */
		public ACMESB() {
			this(0, 0, 0, 0);
		}
		/**
		 * Construct new instance.
		 * @param status status
		 * @param status2 secondary status
		 * @param acmeId ACME ID
		 * @param acmeStatus ACME status
		 */
		public ACMESB(int status, int status2, int acmeId, int acmeStatus) {
			this.status = status;
			this.status2 = status2;
			this.acmeId = acmeId;
			this.acmeStatus = acmeStatus;
		}
		/**
		 * Get status.
		 * @return status
		 */
		public int getStatus() {
			return status;
		}
		/**
		 * Get secondary status.
		 * @return secondary status
		 */
		public int getStatus2() {
			return status2;
		}
		/**
		 * Get ACME ID.
		 * @return ACME ID
		 */
		public int getAcmeId() {
			return acmeId;
		}
		/**
		 * Get ACME status.
		 * @return ACME status
		 */
		public int getAcmeStatus() {
			return acmeStatus;
		}
	}
	/**
	 * Call SYS$ACMW to check username-password.
	 * @param usrnam username
	 * @param pwd password
	 * @return true if valid, false if not valid
	 * @throws UnsupportedEncodingException if username or password can not be encode
	 * @throws RecordException if internal error
	 */
	public static Result<Boolean> SYS$ACMW(String usrnam, String pwd) throws UnsupportedEncodingException, RecordException {
		Item32[] itmarr = new Item32[3];
		itmarr[0] = new Item32LongWordIn(ACME$_LOGON_TYPE, new LongWord(ACME$K_NETWORK));
		itmarr[1] = new Item32CharacterStringIn(ACME$_PRINCIPAL_NAME_IN, new CharacterString(usrnam));
		itmarr[2] = new Item32CharacterStringIn(ACME$_PASSWORD_1, new CharacterString(pwd));
		ItemList32 itmlst = new ItemList32(itmarr);
		Block<ACMESB> sb = new Block<ACMESB>(new ACMESB());
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$ACMW", MISSING,
				                                         pass(new LongWord(ACME$_FC_AUTHENTICATE_PRINCIPAL)).byValue().readOnly(),
				                                         MISSING,
				                                         pass(itmlst).byReference().readOnly(),
				                                         pass(sb).byReference().writeOnly(),
				                                         MISSING,
				                                         MISSING);
		return new Result<Boolean>(sts, isSuccess(sb.getObject(ACMESB.class).getStatus()));
	}
	/**
	 * Call SYS$ASCTIM to get string with current time.
	 * @return return status code + string with time
	 * @throws UnsupportedEncodingException if time string can not be decoded
	 */
	public static Result<String> SYS$ASCTIM() throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(32);
        Word retlen = new Word();
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$ASCTIM", pass(retlen).byReference().writeOnly(),
	               										   pass(buf).byDescriptor().writeOnly(),
	               										   MISSING,
	               										   MISSING);
		return new Result<String>(sts, createString(buf, retlen));
	}
	/**
	 * Call SYS$ASCTIM to get string with specified time.
	 * @param t VMS time
	 * @return return status code + string with time
	 * @throws UnsupportedEncodingException if time string can not be decoded
	 */
	public static Result<String> SYS$ASCTIM(long t) throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(32);
        Word retlen = new Word();
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$ASCTIM", pass(retlen).byReference().writeOnly(),
	               										   pass(buf).byDescriptor().writeOnly(),
	               										   pass(new QuadWord(t)).byReference().readOnly(),
	               										   MISSING);
		return new Result<String>(sts, createString(buf, retlen));
	}
	/**
	 * Call SYS$ASSIGN to open channel to device.
	 * @param devnam device name
	 * @return return status code + channel
	 * @throws UnsupportedEncodingException if device name can not be decoded
	 */
	public static Result<Word> SYS$ASSIGN(String devnam) throws UnsupportedEncodingException {
		Word chan = new Word();
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$ASSIGN", pass(new CharacterString(devnam)).byDescriptor().readOnly(),
				                                           pass(chan).byReference().writeOnly(),
				                                           MISSING,
				                                           MISSING,
				                                           MISSING);
		return new Result<Word>(sts, chan);
	}
	/**
	 * Call SYS$BINTIM to get specified VMS time.
	 * @param timstr time string
	 * @return return status code + VMS time
	 * @throws UnsupportedEncodingException if time string can not be encoded
	 */
	public static Result<Long> SYS$BINTIM(String timstr) throws UnsupportedEncodingException {
		QuadWord tim = new QuadWord();
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$BINTIM", pass(new CharacterString(timstr)).byDescriptor().readOnly(),
				   										   pass(tim).byReference().writeOnly());
		return new Result<Long>(sts, tim.getValue());
	}
	/**
	 * Call LIB$DASSGN to close channel.
	 * @param chan channel
	 * @return return status code
	 */
	public static Result<Void> SYS$DASSGN(Word chan) {
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$DASSGN", pass(chan).byValue().readOnly());
		return new Result<Void>(sts, null);
	}
	/**
	 * Call SYS$GETDVIW to get a list of items.
	 * @param devnam device name
	 * @param itms array of items
	 * @return return status code + Java Map with items
	 * @throws UnsupportedEncodingException if some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<Integer,Object>> SYS$GETDVIW(String devnam, int[] itms) throws UnsupportedEncodingException, RecordException {
		Item32[] itmarr = new Item32[itms.length];
		for(int i = 0; i < itms.length; i++) {
			switch(itms[i]) {
	            case DVI$_DEVTYPE:
	            case DVI$_ERRCNT:
	            case DVI$_MAXBLOCK:
	            case DVI$_FREEBLOCKS:
			    	itmarr[i] = new Item32LongWordOut(itms[i]);
			    	break;
	            case DVI$_VOLNAM:
			    	itmarr[i] = new Item32CharacterStringOut(itms[i], 1024);
			    	break;
			    default:
			    	throw new RuntimeException("Unsupported GETJPI code: " + itms[i]);
			}
		}
		ItemList32 itmlst = new ItemList32(itmarr);
		Block<IOSB> iosb = new Block<IOSB>(new IOSB());
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$GETDVIW", MISSING,
				                                            MISSING,
				                                            pass(new CharacterString(devnam)).byDescriptor().readOnly(),
				                                            pass(itmlst).byReference().readOnly(),
				                                            pass(iosb).byReference().writeOnly(),
				                                            MISSING,
				                                            MISSING,
				                                            MISSING,
				                                            MISSING);
		Map<Integer,Object> res = new HashMap<Integer,Object>();
		for(int i = 0; i < itms.length; i++) {
			switch(itms[i]) {
	            case DVI$_DEVTYPE:
	            case DVI$_ERRCNT:
	            case DVI$_MAXBLOCK:
	            case DVI$_FREEBLOCKS:
			    	LongWord lw = (LongWord)itmarr[i].getBufadr();
			    	res.put(itms[i], lw.getValue());
			    	break;
	            case DVI$_VOLNAM:
			    	CharacterString cs = (CharacterString)itmarr[i].getBufadr();
			    	res.put(itms[i], createTrimmedString(cs));
			    	break;
			    default:
			    	throw new RuntimeException("Unsupported GETDVI code: " + itms[i]);
			}
		}
		return new Result<Map<Integer,Object>>(combineStatus(sts, iosb), res);
	}
	private static Result<Map<Integer,Object>> SYS$GETJPIW(LongWord pid, int[] itms) throws UnsupportedEncodingException, RecordException {
		Item32[] itmarr = new Item32[itms.length];
		for(int i = 0; i < itms.length; i++) {
			switch(itms[i]) {
			    case JPI$_PRI:
			    case JPI$_UIC:
			    case JPI$_STATE:
			    case JPI$_PID:
			    case JPI$_DIRIO:
			    case JPI$_BUFIO:
			    case JPI$_WSSIZE:
			    case JPI$_MODE:
			    	itmarr[i] = new Item32LongWordOut(itms[i]);
			    	break;
			    case JPI$_CURPRIV:
			    case JPI$_CPUTIM:
			    case JPI$_LOGINTIM:
			    	itmarr[i] = new Item32QuadWordOut(itms[i]);
			    	break;
			    case JPI$_PRCNAM:
			    case JPI$_TERMINAL:
			    case JPI$_NODENAME:
			    case JPI$_USERNAME:
			    case JPI$_IMAGNAME:
			    	itmarr[i] = new Item32CharacterStringOut(itms[i], 1024);
			    	break;
			    default:
			    	throw new RuntimeException("Unsupported GETJPI code: " + itms[i]);
			}
		}
		ItemList32 itmlst = new ItemList32(itmarr);
		Block<IOSB> iosb = new Block<IOSB>(new IOSB());
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$GETJPIW", MISSING,
				                                            pass(pid).byReference().readWrite(),
				                                            MISSING,
				                                            pass(itmlst).byReference().readOnly(),
				                                            pass(iosb).byReference().writeOnly(),
				                                            MISSING,
				                                            MISSING);
		Map<Integer,Object> res = new HashMap<Integer,Object>();
		for(int i = 0; i < itms.length; i++) {
			switch(itms[i]) {
			    case JPI$_PRI:
			    case JPI$_UIC:
			    case JPI$_STATE:
			    case JPI$_PID:
			    case JPI$_DIRIO:
			    case JPI$_BUFIO:
			    case JPI$_WSSIZE:
			    case JPI$_MODE:
			    	LongWord lw = (LongWord)itmarr[i].getBufadr();
			    	res.put(itms[i], lw.getValue());
			    	break;
			    case JPI$_CURPRIV:
			    case JPI$_CPUTIM:
			    case JPI$_LOGINTIM:
			    	QuadWord qw = (QuadWord)itmarr[i].getBufadr();
			    	res.put(itms[i], qw.getValue());
			    	break;
			    case JPI$_PRCNAM:
			    case JPI$_TERMINAL:
			    case JPI$_NODENAME:
			    case JPI$_USERNAME:
			    case JPI$_IMAGNAME:
			    	CharacterString cs = (CharacterString)itmarr[i].getBufadr();
			    	res.put(itms[i], createTrimmedString(cs));
			    	break;
			    default:
			    	throw new RuntimeException("Unsupported GETJPI code: " + itms[i]);
			}
		}
		return new Result<Map<Integer,Object>>(combineStatus(sts, iosb), res);
	}
	/**
	 * Call SYS$GETJPIW to get a list of items.
	 * @param pid PID
	 * @param itms array of items
	 * @return return status code + Java Map with items
	 * @throws UnsupportedEncodingException if some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<Integer,Object>> SYS$GETJPIW(int pid, int[] itms) throws UnsupportedEncodingException, RecordException {
		return SYS$GETJPIW(new LongWord(pid), itms);
	}
	/**
	 * Call SYS$GETTIM to get current VMS time.
	 * @return return status code + VMS time
	 */
	public static Result<Long> SYS$GETTIM() {
		QuadWord tim = new QuadWord();
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$GETTIM", pass(tim).byReference().writeOnly(),
				                                           MISSING);
		return new Result<Long>(sts, tim.getValue());
	}
	/**
	 * Call SYS$IDTOASC to get name for identifier.
	 * @param id identifier
	 * @return return status code + identifier name
	 * @throws UnsupportedEncodingException if error decoding identifier name
	 */
	public static Result<String> SYS$IDTOASC(int id) throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(255);
        Word retlen = new Word();
        int sts = call("SYS$PUBLIC_VECTORS", "SYS$IDTOASC", pass(new LongWord(id)).byValue().readOnly(),
                                                            pass(retlen).byReference().writeOnly(),
                                                            pass(buf).byDescriptor().writeOnly(),
                                                            MISSING,
                                                            MISSING,
                                                            MISSING);
        String res = createString(buf, retlen);
        return new Result<String>(sts, res);
	}
	/**
	 * Class ID_CTX represents the context for SYS$IDTOASC calls.
	 */
	public static class ID_CTX {
		private LongWord ctx = new LongWord();
		/**
		 * Call SYS$IDTOASC to return next identifier.
		 * @return return status code + identifier
		 * @throws UnsupportedEncodingException if error decoding identifier name
		 */
		public Result<Pair<Integer, String>> SYS$IDTOASC() throws UnsupportedEncodingException {
	        CharacterString buf = new CharacterString(255);
	        Word retlen = new Word();
	        LongWord resid = new LongWord();
	        int sts = call("SYS$PUBLIC_VECTORS", "SYS$IDTOASC", pass(new LongWord(-1)).byValue().readOnly(),
	                                                            pass(retlen).byReference().writeOnly(),
	                                                            pass(buf).byDescriptor().writeOnly(),
	                                                            pass(resid).byReference().writeOnly(),
	                                                            MISSING,
	                                                            pass(ctx).byReference().readWrite());
	        String res = createString(buf, retlen);
	        return new Result<Pair<Integer,String>>(sts, new Pair<Integer,String>(resid.getValue(), res));
		}
	}
	/**
	 * Call SYS$GETUAI to get a list of items.
	 * @param usrnam username
	 * @param itms array of items
	 * @return return status code +  Java Map with items
	 * @throws UnsupportedEncodingException if some string values can not be decoded
	 * @throws RecordException if internal error
	 */
	public static Result<Map<Integer,Object>> SYS$GETUAI(String usrnam, int[] itms) throws UnsupportedEncodingException, RecordException {
		Item32[] itmarr = new Item32[itms.length];
		for(int i = 0; i < itms.length; i++) {
			switch(itms[i]) {
				case UAI$_UIC:
				case UAI$_LOGFAILS:
				case UAI$_SALT:
				case UAI$_ENCRYPT:
				case UAI$_PRI:
				case UAI$_MAXJOBS:
				case UAI$_PRCCNT:
				case UAI$_WSQUOTA:
				case UAI$_WSEXTENT:
				case UAI$_PGFLQUOTA:
			    	itmarr[i] = new Item32LongWordOut(itms[i]);
			    	break;
				case UAI$_PWD:
				case UAI$_PWD_DATE:
				case UAI$_PRIV:
				case UAI$_DEF_PRIV:
			    	itmarr[i] = new Item32QuadWordOut(itms[i]);
			    	break;
				case UAI$_USERNAME:
				case UAI$_ACCOUNT:
				case UAI$_OWNER:
				case UAI$_DEFDEV:
				case UAI$_DEFDIR:
			    	itmarr[i] = new Item32CharacterStringOut(itms[i], 1024);
			    	break;
			    default:
			    	throw new RuntimeException("Unsupported GETJPI code: " + itms[i]);
			}
		}
		ItemList32 itmlst = new ItemList32(itmarr);
		int sts = call("SECURESHRP", "SYS$GETUAI", MISSING,
				                                   MISSING,
				                                   pass(new CharacterString(usrnam)).byDescriptor().readOnly(),
			                                       pass(itmlst).byReference().readOnly(),
			                                       MISSING,
			                                       MISSING,
				                                   MISSING);
		Map<Integer,Object> res = new HashMap<Integer,Object>();
		for(int i = 0; i < itms.length; i++) {
			switch(itms[i]) {
				case UAI$_UIC:
				case UAI$_LOGFAILS:
				case UAI$_SALT:
				case UAI$_ENCRYPT:
				case UAI$_PRI:
				case UAI$_MAXJOBS:
				case UAI$_PRCCNT:
				case UAI$_WSQUOTA:
				case UAI$_WSEXTENT:
				case UAI$_PGFLQUOTA:
			    	LongWord lw = (LongWord)itmarr[i].getBufadr();
			    	res.put(itms[i], lw.getValue());
			    	break;
				case UAI$_PWD:
				case UAI$_PWD_DATE:
				case UAI$_PRIV:
				case UAI$_DEF_PRIV:
			    	QuadWord qw = (QuadWord)itmarr[i].getBufadr();
			    	res.put(itms[i], qw.getValue());
			    	break;
				case UAI$_USERNAME:
				case UAI$_ACCOUNT:
				case UAI$_OWNER:
				case UAI$_DEFDEV:
				case UAI$_DEFDIR:
			    	CharacterString cs = (CharacterString)itmarr[i].getBufadr();
			    	res.put(itms[i], createTrimmedString(cs));
			    	break;
			    default:
			    	throw new RuntimeException("Unsupported GETUAI code: " + itms[i]);
			}
		}
		return new Result<Map<Integer,Object>>(sts, res);
	}
	/**
	 * Class PID_CTX represents the context for SYS$PROCESS_SCAN calls.
	 */
	public static class PID_CTX {
		private LongWord ctx = new LongWord();
		/**
		 * Call SYS$PROCESS_SCAN to find next matching PID.
		 * @param code code for SYS$PROCESS_SCAN
		 * @param value value for SYS$PROCESS_SCAN
		 * @return return status code
		 * @throws UnsupportedEncodingException if value can not be encoded
		 */
		public Result<Void> SYS$PROCESS_SCAN(int code, String value) throws UnsupportedEncodingException {
			Result<Integer> res = SYS.SYS$PROCESS_SCAN(code, value);
			this.ctx = new LongWord(res.getResult());
			return new Result<Void>(res.getStatus(), null);
		}
		/**
		 * Call SYS$GETJPIW to get lost of items.
		 * @param itms array of items
		 * @return return status code + Java Map of items
		 * @throws UnsupportedEncodingException if some string values can not be decoded
		 * @throws RecordException if internal error
		 */
		public Result<Map<Integer,Object>> SYS$GETJPIW(int[] itms) throws UnsupportedEncodingException, RecordException {
			return SYS.SYS$GETJPIW(ctx, itms);
		}
	}
	/**
	 * Call SYS$PROCESS_SCAN to get first matching PID.
	 * @param code code for SYS$PROCESS_SCAN
	 * @param value value for SYS$PROCESS_SCAN
	 * @return return status code + PID
	 * @throws UnsupportedEncodingException if value can not be encoded
	 */
	public static Result<Integer> SYS$PROCESS_SCAN(int code, String value) throws UnsupportedEncodingException {
		LongWord cntx = new LongWord();
		ItemList32 itmlst = new ItemList32(new Item32[] { new Item32CharacterStringIn(code, new CharacterString(value), value.contains("*") ? PSCAN$M_WILDCARD : 0) });
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$PROCESS_SCAN", pass(cntx).byReference().readWrite(),
				                                                 pass(itmlst).byReference().readOnly());
		return new Result<Integer>(sts, cntx.getValue());
	}
	/**
	 * Call SYS$QIOW.
	 * @param chan channel
	 * @param func function
	 * @param iosb IOSB
	 * @param p1 p1
	 * @param p2 p2
	 * @return return status code
	 * @throws RecordException if internal error
	 * @throws UnsupportedEncodingException if something can not be encoded or decoded
	 */
	public static Result<Void> SYS$QIOW(Word chan, int func, Block<IOSB> iosb, TypeWrapper<byte[]> p1, TypeWrapper<Integer> p2) throws RecordException, UnsupportedEncodingException {
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$QIOW", MISSING,
								                		 pass(chan).byValue().readOnly(),
								                		 pass(new LongWord(func)).byValue().readOnly(),
								                		 pass(iosb).byReference().writeOnly(),
								                		 MISSING,
								                		 MISSING,
								                		 pass(p1).byReference().readWrite(),
								                		 pass(p2).byValue().readOnly(),
								                		 MISSING,
								                		 MISSING,
								                		 MISSING,
								                		 MISSING);
		return new Result<Void>(combineStatus(sts, iosb), null);
	}
	/**
	 * Call SYS$SETDDIR to get current default directory.
	 * @return return status code + current default directory
	 * @throws UnsupportedEncodingException if directory can not be decoded
	 */
	public static Result<String> SYS$SETDDIR() throws UnsupportedEncodingException {
        CharacterString buf = new CharacterString(1024);
        Word retlen = new Word();
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$SETDDIR", MISSING,
	               											pass(retlen).byReference().writeOnly(),
	               											pass(buf).byDescriptor().readWrite());
		return new Result<String>(sts, createString(buf, retlen));
	}
	/**
	 * Call SYS$SETDDIR to change current default directory.
	 * @param newdir new default directory
	 * @return return status code
	 * @throws UnsupportedEncodingException if directory can not be encoded
	 */
	public static Result<Void> SYS$SETDDIR(String newdir) throws UnsupportedEncodingException {
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$SETDDIR", pass(new CharacterString(newdir)).byDescriptor().readOnly(),
	               											MISSING,
	               											MISSING);
		return new Result<Void>(sts, null);
	}
	/**
	 * Class TERMINAL_IO represents the context (actually channel) for SYS$ASSIGN, SYS$QIOW and SYSDASSGN calls for terminal IO.
	 */
	public static class TERMINAL_IO {
		@Struct(alignment=Alignment.PACKED)
		public static class CharBufTypAhd {
			@StructField(n=0,type=FieldType.INT2)
			private int noChar;
			@StructField(n=1,type=FieldType.INT1)
			private int firstChar;
			@StructField(n=2,type=FieldType.INT1)
			private int fill1;
			@StructField(n=3,type=FieldType.INT4)
			private int fill2;
			public int getNoChar() {
				return noChar;
			}
			public int getFirstChar() {
				return firstChar;
			}
		}
		private Word chan;
		/**
		 * Call SYS$ASSIGN to open channel to default terminal (TT:).
		 * @return return status code
		 * @throws UnsupportedEncodingException if device name can not be decoded
		 */
		public Result<Void> SYS$ASSIGN() throws UnsupportedEncodingException {
			return SYS$ASSIGN("TT:");
		}
		/**
		 * Call SYS$ASSIGN to open channel to specified terminal.
		 * @param devnam terminal name
		 * @return return status code
		 * @throws UnsupportedEncodingException if device name can not be decoded
		 */
		public Result<Void> SYS$ASSIGN(String devnam) throws UnsupportedEncodingException {
			 Result<Word> res = SYS.SYS$ASSIGN(devnam);
			 chan = res.getResult();
			 return new Result<Void>(res.getStatus(), null);
		}
		/**
		 * Call SYS$QIOW to write string to terminal (puts functionality).
		 * @param s string to write
		 * @return return status code
		 * @throws RecordException if internal error
		 * @throws UnsupportedEncodingException if string can not be encoded
		 */
		public Result<Void> SYS$QIOW_puts(String s) throws RecordException, UnsupportedEncodingException {
			Block<IOSB> iosb = new Block<IOSB>(new IOSB());
			return SYS.SYS$QIOW(chan, IO$_WRITEVBLK, iosb, new ByteBuffer(s), new LongWord(s.length()));
		}
		/**
		 * Call SYS$QIOW to read line from terminal (gets functionality).
		 * @return return status code + string read 
		 * @throws RecordException if internal error
		 * @throws UnsupportedEncodingException if string can not be decoded
		 */
		public Result<String> SYS$QIOW_gets() throws RecordException, UnsupportedEncodingException {
			return SYS$QIOW_gets(true);
		}
		/**
		 * Call SYS$QIOW to read line from terminal (gets functionality).
		 * @param echo echo input
		 * @return return status code + string read 
		 * @throws RecordException if internal error
		 * @throws UnsupportedEncodingException if string can not be decoded
		 */
		public Result<String> SYS$QIOW_gets(boolean echo) throws RecordException, UnsupportedEncodingException {
			Block<IOSB> iosb = new Block<IOSB>(new IOSB());
			ByteBuffer buf = new ByteBuffer(4096);
			Result<Void> res = SYS.SYS$QIOW(chan, IO$_READVBLK|(echo ? 0 : IO$M_NOECHO), iosb, buf, new LongWord(buf.getLength()));
			return new Result<String>(res.getStatus(), createString(buf, iosb.getObject(IOSB.class).getCount()));
		}
		/**
		 * Call SYS$QIOW to read single character from terminal (getch functionality).
		 * @return return status code + character read 
		 * @throws RecordException if internal error
		 * @throws UnsupportedEncodingException will never happen
		 */
		public Result<Character> SYS$QIOW_getch() throws RecordException, UnsupportedEncodingException {
			Block<IOSB> iosb = new Block<IOSB>(new IOSB());
			ByteBuffer buf = new ByteBuffer(1);
			Result<Void> res = SYS.SYS$QIOW(chan, IO$_READVBLK|IO$M_NOECHO|IO$M_NOFILTR, iosb, buf, new LongWord(buf.getLength()));
			return new Result<Character>(res.getStatus(), (char)buf.getValue()[0]);
		}
		/**
		 * Call SYS$QIOW to check if there is input available to be read (kbhit functionality).
		 * @return  return status code + indication if input available
		 * @throws RecordException if internal error
		 * @throws UnsupportedEncodingException  will never happen
		 */
		public Result<Boolean> SYS$QIOW_kbhit() throws RecordException, UnsupportedEncodingException {
			Block<IOSB> iosb = new Block<IOSB>(new IOSB());
			Block<CharBufTypAhd> cb = new Block<CharBufTypAhd>(new CharBufTypAhd());
			Result<Void> res = SYS.SYS$QIOW(chan, IO$_SENSEMODE|IO$M_TYPEAHDCNT, iosb, cb, new LongWord(cb.getLength()));
			return new Result<Boolean>(res.getStatus(), cb.getObject(CharBufTypAhd.class).getNoChar() > 0);
		}
		/**
		 * Call LIB$DASSGN to close channel.
		 * @return return status code
		 */
		public Result<Void> SYS$DASSGN() {
			return SYS.SYS$DASSGN(chan);
		}
	}
	/**
	 * Call SYS$TRNLNM to translate logical name.
	 * @param tblnam table name
	 * @param lognam logical name
	 * @return return status code + translate logical name
	 * @throws UnsupportedEncodingException if table or logical name can not be encoded or translation can not be decoded
	 */
	public static Result<String> SYS$TRNLNM(String tblnam, String lognam) throws UnsupportedEncodingException {
		Item32[] itmarr = new Item32[0];
    	itmarr[0] = new Item32CharacterStringOut(LNM$_STRING, 256);
		ItemList32 itmlst = new ItemList32(itmarr);
		int sts = call("SYS$PUBLIC_VECTORS", "SYS$TRNLNM", MISSING,
											               pass(new CharacterString(tblnam)).byDescriptor().readOnly(),
											               pass(new CharacterString(lognam)).byDescriptor().readOnly(),
											               MISSING,
											               pass(itmlst).byReference().readOnly());
		return new Result<String>(sts, createTrimmedString((CharacterString)itmarr[0].getBufadr()));
	}
}
